/*********************************************************
 *
 * IMS_MIDI.c
 *
 * Authors: Ben McQuiston
 *
 * C function to convert our house IMS files into a usable MIDI file
 * 
 * Last Revision: 
 *       - October 25, 2006
 *			Major structure created for Mthd print and 
 *          first MTrk with tempo specifiers 
 *
 *       - October 29, 2006
 *			Note processing code added, to decide how many bytes
 *			a given track needs to be, and write the actual delta
 *			times and note values to the MIDI file.
 *  
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FileFunctions.h"

int
ImsToMIDI(void)
{
	//printf("Generating MIDI file...\n");
	
	extern unsigned int * pex_IMS;
	extern unsigned int * pex_MIDI;
	// extern unsigned int IMS;
	extern unsigned int IMS[100];
	extern unsigned int MIDI[1000];
	extern int IMSSz;
	extern int MIDISz;
	
	int checkBuff[50] ={0};
	char notes[200] = {0};
	int temp;
    int i = 0, j = 0;
	int tempo;
	int divisor = 60000000;
	int mspqn;

	int fileMrk;
	int numTrkByts = 0;
	int ctr = 0;

	//FILE *fp1;
	//FILE *fp2;

	char ch;
	char keySig;
	char timeSigTop;
	char timeSigBot;
	char instr;
	char noteVal;
	
	
//	fp1 = fopen(filename, "rb");
//	fp2 = fopen("output.mid","w");

	//if(fp1 == NULL){
//	  printf("ERROR\n");
//	  return -1;
//	}

  pex_MIDI = MIDI;

/********************************************************
*
*  Mthd header to file...this will always be the same
*
*********************************************************/


  ExMIDIWrite(0x4d); //fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x68,0x64);
  ExMIDIWrite(0x54);
  ExMIDIWrite(0x68);             /* Mthd */
  ExMIDIWrite(0x64);
  
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(0x06);	//fprintf(fp2,"%c",0x06);/* Always 6 bytes in Mthd header */
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(0x01);	//fprintf(fp2,"%c",0x01);/* MIDI Type 1 */
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(0x02);	//fprintf(fp2,"%c",0x02);/* Number of Tracks... could change?*/
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(0xF0);	//fprintf(fp2,"%c",0xF0); /*240 PPQN always..so dont forget*/

	 /*fprintf(fp2,"%c",zero); delta time 00 */

/********************************************************
*  
*   Mtrk Header...
*
*********************************************************/

  ExMIDIWrite(0x4d); //fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x72,0x6B); 
  ExMIDIWrite(0x54);
  ExMIDIWrite(0x72);             /* Mtrk */
  ExMIDIWrite(0x6B);
  
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
  ExMIDIWrite(0x19);	//fprintf(fp2,"%c",0x19); /* 25 bytes in first track...stays the same */

  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero); /* delta time 00 */
	
/******************************************************
*
*  Start at beginning of IMS file to start getting 
*  musical information...
*
********************************************************/

  pex_IMS = IMS;	//fseek(fp1,0,SEEK_SET);
  
	
	/* tempo cammand FF 51 03 xx xx xx = micros / quarter note */

    read_extmem(&temp, pex_IMS, 1);	//fscanf(fp1,"%c",&ch);
	tempo = (temp & 0xFF000000) >> 24;//tempo = (int) ch; 
	//tempo = tempo & ~(0xFFFFFFFF << 8);mask off all but the lower byte); 

	mspqn = divisor / tempo;
	
  ExMIDIWrite(0xFF);	//fprintf(fp2,"%c%c%c", 0xFF,0x51,0x03); tempo command 
  ExMIDIWrite(0x51);
  ExMIDIWrite(0x03);
/*  
ch =(char) ((mspqn  & 0x00FF0000)>>16);
ExIMSWrite(ch);
ch = (mspqn  & 0x0000FF00)>>8;
ExIMSWrite(ch);
ch = (mspqn  & 0x000000FF);
ExIMSWrite(ch);
*/
  ExMIDIWrite((char) ((mspqn  & 0x00FF0000)>>16));//fprintf(fp2,"%c%c%c",*((char *) &mspqn +2),*((char *) &mspqn+1), *((char *) &mspqn));
  ExMIDIWrite((char) ((mspqn  & 0x0000FF00)>>8));
  ExMIDIWrite((char) (mspqn  & 0x000000FF));

  ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);  /*delta time 00 */

	/* key sig command FF 59 02 xx xx = flats or sharps and major or minor */

	//fscanf(fp1,"%c",&ch);

	//keySig = (temp & 0x00FF0000) >> 16;   //keySig = ch;
	
	/* logic........for now we are just going to force it 
	* to C major no flats we are not sure why we need this...*/
	
	ExMIDIWrite(0xFF); //fprintf(fp2,"%c%c%c",0xFF,0x59,0x02);
	ExMIDIWrite(0x59);
	ExMIDIWrite(0x02);
	
	ExMIDIWrite(ZERO); //fprintf(fp2,"%c",zero); /* no flats (C) */
	ExMIDIWrite(ZERO); //fprintf(fp2,"%c",zero); /* major */ 
	ExMIDIWrite(ZERO); //fprintf(fp2,"%c",zero); /* delta time 00 */

	/* Time Signature command FF 58 04 xx xx xx xx 
	* numerator and denominator of time signature
	* number of MIDI clocks per beat (always 24)
	* number of 32nd notes per quater note (always 8)
	*/
	
	//fscanf(fp1,"%c",&ch);
	timeSigTop = (temp & 0x0000FF00) >> 8; //timeSigTop = ch;
	
	ExMIDIWrite(0xFF);    //fprintf(fp2,"%c%c%c",0xFF,0x58,0x04);
	ExMIDIWrite(0x58);
	ExMIDIWrite(0x04);

	ExMIDIWrite(timeSigTop);    //fprintf(fp2,"%c",ch);

    timeSigBot = (temp & 0x000000FF);    //fscanf(fp1,"%c",&ch);
	ch = timeSigBot;
	
    /* denominator must be a negative power of four,
	*  so we check if its a 4 or an 8 and print the corresponding
	* power to the file...
	*/

	if ((int) ch == 4){
		ExMIDIWrite(0x02); //fprintf(fp2,"%c",0x02); 
	}
	else if ((int) ch == 8){
		ExMIDIWrite(0x03); //fprintf(fp2,"%c",0x03);
	}

	ExMIDIWrite(0x18); //fprintf(fp2,"%c%c",0x18,0x08); 
	ExMIDIWrite(0x08); //24 MIDI clocks per beat, 8 32nds per quarter*/								   

	ExMIDIWrite(ZERO); //fprintf(fp2,"%c",zero);  /* delta time 00 */

	ExMIDIWrite(0xFF);  //fprintf(fp2,"%c%c%c",0xFF,0x2F,0); /* End of Track Marker */
    ExMIDIWrite(0x2F);
    ExMIDIWrite(ZERO);
    
    //MIDIBuffFlush();

/*************************************************************
*
*  NOTE PROCESSING
*
*  Actually deal with the note values and lengths
*  Figure out how long it is going to be so that the Mtrk
*  header can tell how many bytes to expect, then just 
*  process all of the notes two bytes at a time...
*
************************************************************/


	ExMIDIWrite(0x4D);  //fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x72,0x6B); /* Mtrk */
	ExMIDIWrite(0x54);
	ExMIDIWrite(0x72);
	ExMIDIWrite(0x6B);
	
	/* So to make this work we have to do an initial search through
	* the file to figure out what is in there...basically to figure out
	* out how many bytes are goiung to be in the file....*/
	
	//This is going to be alot easier, just check the IMSSz variable to figure out
	// how big the IMS file is, subtract 6 for the header, then divide by two for 
	// the number of notes, and multiply by nine for the number of bytes in each 
	//note on, note off sequence.
	
	read_extmem(&temp, (pex_IMS + 4), 1);	//fscanf(fp1,"%c",&ch);
	ch = (temp & 0xFF000000) >> 24;
	instr = ch;
	
	fileMrk = 5; 
/*
	while (fscanf(fp1,"%c",&ch) == 1){
		
		fscanf(fp1,"%c",&ch);
		if (ch == 0x03 || ch == 0x04 || ch == 0x05){
			numTrkByts += 8;
		}
		else{
			numTrkByts += 9;
		}
	}
*/
    numTrkByts = (((IMSSz - 6) / 2) * 9);	
	
    //printf("count: %d\n", numTrkByts);

	//fseek(fp1,fileMrk,SEEK_SET);

	numTrkByts += 8; /* + 4 to account for the four bytes to tell
					 * what instrument is to be played and +4 for
					 the end of track bytes*/

//	printf("numTrkByts: %d\n",numTrkByts);
  ExMIDIWrite((char) ((numTrkByts  & 0x00FF0000)>>24));//fprintf(fp2,"%c",*((char *) &numTrkByts + 3)); 
  ExMIDIWrite((char) ((numTrkByts  & 0x0000FF00)>>16));//fprintf(fp2,"%c",*((char *) &numTrkByts + 2));
  ExMIDIWrite((char) ((numTrkByts  & 0x000000FF)>>8));//fprintf(fp2,"%c",*((char *) &numTrkByts + 1));
  ExMIDIWrite((char) (numTrkByts  & 0x000000FF));	//fprintf(fp2,"%c",*((char *) &numTrkByts));


		/* actual number of bytes in the coming track */
	
	/*After we figure out how many bytes there will be in the track,
	* we have to actually process the notes and things...*/

	ExMIDIWrite(ZERO);  //fprintf(fp2,"%c",zero);
	ExMIDIWrite(0xC0);  //fprintf(fp2,"%c%c",0xC0,instr); /*tells what instrument it is ..*/
	ExMIDIWrite(instr);
	ExMIDIWrite(ZERO);	//fprintf(fp2,"%c",zero);
	   
	notes[0] = (temp & 0x0000FF00) >> 8;
	notes[1] = (temp & 0x000000FF);	
	ctr = 2;
	
	for (i = 0; i <= (IMSSz - 8) / 4; i++){
	    //printf("i: %d\n",i);
	    read_extmem(&temp, (pex_IMS + 8 + (4*i)), 1);
	   // printf("temp: %x\n",temp);
	    for (j = 0; j < 4; j++){
	        notes[j + ctr] = ((temp >> (24 - (8 * j))) & 0x000000FF);
	    }
	    ctr = ctr + 4;  
	}
	   
//	for (i = 0; i < (IMSSz - 6); i++){
//	    printf("notes[%d] = %x\n",i,notes[i]);
//	}
	
	for (i = 0; i < (IMSSz - 6); i++){	
		if (i % 2 == 0){
	      noteVal = notes[i];	//while (fscanf(fp1,"%c",&ch) == 1){
		  //noteVal = ch;
		  ExMIDIWrite(0x90);		//fprintf(fp2,"%c%c%c",0x90,noteVal,0x40);
		  ExMIDIWrite(noteVal);
		  ExMIDIWrite(0x40);
	    }
	    else {
		  ch = notes[i];    //fscanf(fp1,"%c",&ch);
		  //printf("ch: %X\n",ch);
		  if (ch == 0){ /* whole note = 960 pulses */
			ExMIDIWrite(0x87);    //fprintf(fp2,"%c",0x87);
			ExMIDIWrite(0x40);    //fprintf(fp2,"%c",0x40);
		  }
		  else if (ch == 0x01){ /* half note = 480 pulses */
			ExMIDIWrite(0x83);   // fprintf(fp2,"%c",0x83);
		  	ExMIDIWrite(0x60);   // fprintf(fp2,"%c",0x60);
		  }
		  else if (ch == 0x02){ /* quarter note = 240 pulses */
			ExMIDIWrite(0x81);   // fprintf(fp2,"%c",0x81);
			ExMIDIWrite(0x70);   // fprintf(fp2,"%c",0x70);
		  }
		  else if (ch == 0x03){ /* eigth note = 120 pulses */
			ExMIDIWrite(0x78);   // fprintf(fp2,"%c",0x78);
		  }
		  else if (ch == 0x04){ /* 16th note = 60 pulses */
			ExMIDIWrite(0x3C);   // fprintf(fp2,"%c",0x3C);
		  }
		  else if (ch == 0x05){/* 32nd note = 30 pulses */
			ExMIDIWrite(0x1E);   // fprintf(fp2,"%c",0x1E);
		  }
		  else if (ch == 0x06){/* dotted eigth note = 180 pulses */
			ExMIDIWrite(0x81);  //  fprintf(fp2,"%c",0x81);
			ExMIDIWrite(0x34);  //  fprintf(fp2,"%c",0x34);
		  }
		  else if (ch == 0x07){ /* dotted quarter note = 360 pulses */
			ExMIDIWrite(0x82);   // fprintf(fp2,"%c",0x82);
			ExMIDIWrite(0x68);    //fprintf(fp2,"%c",0x68);
		  }
		  else if (ch == 0x08){ /* dotted half note = 720 pulses */
			ExMIDIWrite(0x83);    //fprintf(fp2,"%c",0x83);
		    ExMIDIWrite(0x50);    //	fprintf(fp2,"%c",0x50);
		  }
		
		  ExMIDIWrite(0x80);    //fprintf(fp2,"%c%c%c",0x80,noteVal,0x40);
		  ExMIDIWrite(noteVal);    
		  ExMIDIWrite(0x40);    		
	    
	      if (i == (IMSSz - 7)){
	        //printf("hello\n");
	        ExMIDIWrite(0xFF);
	        ExMIDIWrite(0x7F);
	        ExMIDIWrite(0xFF);
	        ExMIDIWrite(0x2F);
	        ExMIDIWrite(ZERO);
		  }
		  else {
		    ExMIDIWrite(ZERO);
		  }
		  
		}

	}
	//fseek(fp2,-1,SEEK_CUR);

//	fprintf(fp2,"%c",0xFF);
//	fprintf(fp2,"%c",0x7F);
//	fprintf(fp2,"%c%c%c",0xFF,0x2F,0); /* End of Track Marker */	

////////////////////////////////
//
//  Check whats in buffer...
//	
/////////////////////////////
/*
  MIDIBuffFlush();
  
      read_extmem(checkBuff,MIDI,32);	
    
    for (i = 0; i < 32; i++){
        printf("%8X\n",checkBuff[i]);
    }

*/
/////////////////////////////////////////
	
    //printf("MIDIsize: %d\n",MIDISz);
	return 0;
	
}
